import uuid
from enum import Enum
from typing import Any, Dict, List, Optional

from pydantic import BaseModel, ConfigDict, Field, field_validator
from rfc3339_validator import validate_rfc3339
from typing_extensions import Annotated


class Job(BaseModel, extra="forbid"):
    jid: str = Field(default_factory=lambda: uuid.uuid4().hex)
    jobtype: str
    args: List[Any]
    queue: str = "default"
    reserve_for: Annotated[int, Field(ge=60)] = 1800
    at: str = ""
    retry: Annotated[int, Field(ge=-1)] = 25
    backtrace: Annotated[int, Field(ge=0)] = 5
    custom: Optional[Dict] = None

    @field_validator("at")
    @classmethod
    def at_is_rfc3339(cls, at):
        del cls
        if len(at) > 0 and not validate_rfc3339(at):
            raise ValueError(f"{at} is not RFC3339 valid")
        return at


class TargetJob(BaseModel, extra="forbid"):
    jobtype: str
    args: List[Any]
    queue: str = "default"


class Batch(BaseModel, extra="forbid"):
    parent_bid: Optional[str]
    description: Optional[str]
    success: Optional[TargetJob]
    complete: Optional[TargetJob]


class JobFilter(BaseModel, extra="forbid"):
    jids: Optional[List[str]]
    regexp: Optional[str]
    jobtype: Optional[str]


class Cmd(str, Enum):
    clear = "clear"
    kill = "kill"
    discard = "discard"
    requeue = "requeue"


class Target(str, Enum):
    retries = "retries"
    scheduled = "scheduled"
    dead = "dead"


class MutateOperation(BaseModel):
    model_config = ConfigDict(use_enum_values=True, extra="forbid")

    cmd: Cmd
    target: Target
    filter: Optional[JobFilter] = None



import json
from typing import Dict, List

from .client import Client
from .models import Batch, Job
from .util import helper


class Producer:
    """
    Faktory Producer.

    Parameters
    ----------
    client : pyfaktory.Client
        Faktory client whose `role` must be either 'producer' or 'both'.
    """

    def __init__(self, client: Client) -> None:
        if client.role == "consumer":
            raise ValueError(
                "Provided client is exclusively consumer and can't act as a producer"
            )
        self.client = client

    def push(self, job: Job) -> bool:
        return self.client._push(job.model_dump(exclude_none=True))

    def push_bulk(self, jobs: List[Job]) -> Dict:
        msg = self.client._pushb([j.model_dump(exclude_none=True) for j in jobs])
        _, data = helper.RESP.parse_bulk_string(msg)
        return json.loads(data)

    def batch_new(self, batch: Batch) -> bool:
        return self.client._batch_new(batch.model_dump(exclude_none=True))

    def batch_open(self, bid: str) -> bool:
        return self.client._batch_open(bid)

    def batch_commit(self, bid: str) -> bool:
        return self.client._batch_commit(bid)



####
## Client
####

CRLF = "\r\n"
DEFAULT_FAKTORY_URL = "tcp://localhost:7419"
DEFAULT_PORT = 7419
DEFAULT_LABELS = ["python"]
DEFAULT_QUEUES = ["default"]

####
## Producer
####

RECOMMENDED_BEAT_PERIOD = 15  # seconds
MIN_ALLOOWABLE_BEAT_PERIOD = 5  # seconds
MAX_ALLOOWABLE_BEAT_PERIOD = 60  # seconds

####
## Consumer
####

DEFAULT_GRACE_PERIOD = 25
MAX_GRACE_PERIOD = 30
